__global__ static void 
#ifdef HST_SET_BOUNDS
__launch_bounds__(256, HST_NEWTEX_MIN_BLOCKS)
#endif /* HST_SET_BOUNDS */
hst_kepler_orig_kernel(cudaTextureObject_t texptr, int num_proj, int num_bins, vfloat *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    vfloat res[4][4] = {0};

#if  (defined(HST_SHFL_SUM) && (SLICE_BLOCK == 1))
    __shared__ vfloat buf[16][20]; 
    __shared__ vfloat fin[16][16];
#else
    __shared__ vfloat buf[16][16]; 
    __shared__ vfloat fin[16][16];
#endif


    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    const int sidx = tidx % 4;
    const int sidy = tidx / 4;

    const float x = bidx + sidx + apos_off_x;
    const float y = bidy + sidy + apos_off_y;

    const float projf = tidy + 0.5f;

    for (int proje=0; proje<num_proj; proje+=16) {
        const float4 all = c_all[proje+tidy];
        const float h = all.z + x * all.x - y * all.y;

#pragma unroll 4
        for (int i = 0; i < 4; i++) {
#pragma unroll 4
	    for (int j = 0; j < 4; j++) {
	        float subh = h + 4 * j * all.x - 4 * i * all.y;
	        res[i][j] += hst_tex(texptr, subh, proje + projf);
	    }
        }

    }


#pragma unroll 4
    for (int i = 0; i < 4; i++) {
#pragma unroll 4
	for (int j = 0; j < 4; j++) {
            buf[tidx][tidy] = res[i][j];

            __syncthreads();


#ifdef HST_SHFL_SUM
            vfloat val = buf[tidy][tidx];
            for (int k=16; k>=1; k/=2)
        	shfl_sum(val, k);
# else // HST_SHFL_SUM
            for (int k=8; k>=1; k/=2) {
                buf[tidy][tidx] += buf[tidy][tidx+k];
            	    // We need it instead of volatile. Could be faster if buf declared volatile
                __syncthreads();
            }
            vfloat val = buf[tidy][0];
#endif //  HST_SHFL_SUM

            const int rx = 4 * j + tidy%4;
            const int ry = 4 * i + tidy/4;

            if (!tidx) {
                fin[ry][rx] = val;
            }

            __syncthreads();
        }
    }


    const int idx = bidx + tidx;
    const int idy = bidy + tidy;

    d_SLICE[BLOCK_SIZE_X * gridDim.x * idy + idx] = fin[tidy][tidx];
}

/*
#ifdef SIMPLE_SHMEM
# define GET_CACHE_SIMD(v1, v2, vc, i) ({ int bin = (int)minh.vc; v1.vc = CACHE(i, bin, 0); v2.vc = CACHE(i, bin, 1); })
#else // SIMPLE_SHMEM
# define GET_CACHE_SIMD(v1, v2, vc, i) ({ float2 c = cache[i][(int)minh.vc]; v1.vc = c.x; v2.vc = c.y; })
#endif // SIMPLE_SHMEM 

//__global__
__device__ static void hst_cuda_simd_linear_kernel(const float * __restrict__ sino, int num_proj, int num_bins, int bin_pitch, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float4 res = {0.f, 0.f, 0.f, 0.f};
    
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = 2 * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + 2 * blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + 1 * tidx; //2
    const int idy = bidy + 1 * tidy; //2

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = idx + apos_off_x;
    const float y = idy + apos_off_y;

    __shared__  float2   cache_subh[BLOCK_SIZE_Y][BLOCK_SIZE_X];
    __shared__  float   cache_sin[BLOCK_SIZE_Y];

#ifdef HST_TUNE_SHMEM
    __shared__  float   cache[2][BLOCK_SIZE_Y + 1][(3 * BLOCK_SIZE_X)];
#else 
    __shared__  float2   cache[BLOCK_SIZE_Y][3 * BLOCK_SIZE_X + 1];
#endif

//#pragma unroll 8



    HST_FETCH_VAR(projf, tidy);
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE_Y) {
        float4 all = c_all[proje + tidy];
	float minh = floor(all.z + bx * all.x - by * all.y + all.w);

	float subh = (all.z + x * all.x - minh);
	cache_subh[tidy][tidx].x = subh;
	cache_subh[tidy][tidx].y = subh + BLOCK_SIZE_X * all.x;
	if (tidx == 0) cache_sin[tidy] = all.y;


	HST_FETCH_VAR(binf, minh);
#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;

	    float v = GET_SINO(binf + pos, proje + projf);
	    CACHE(tidy, pos, 0) = v;
	}
	__syncthreads();	// We may replace __syncthreads() with volatile declaration of shmem cache

#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;
	    CACHE(tidy, pos, 1) = CACHE(tidy, pos + 1, 0) - CACHE(tidy, pos, 0);
	}
	__syncthreads();


#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE_Y; i++) {
	    float4 v1, v2;

	    float2 subh = cache_subh[i][tidx];
//	    float csin = c_all[proje + i].y;
	    float csin = cache_sin[i];

	    float4 minh = {
		subh.x - y * csin,
		subh.y - y * csin,
		subh.x - (y + BLOCK_SIZE_Y) * csin,
		subh.y - (y + BLOCK_SIZE_Y) * csin
	    };

	    GET_CACHE_SIMD(v1, v2, x, i);
	    GET_CACHE_SIMD(v1, v2, y, i);
	    GET_CACHE_SIMD(v1, v2, z, i);
	    GET_CACHE_SIMD(v1, v2, w, i);

	    res += v1 + (minh - floorf(minh)) * v2;
	}

	__syncthreads();

    }

    d_SLICE[BLOCK_SIZE_X * 2 * gridDim.x * (idy               ) + idx               ] = res.x;
    d_SLICE[BLOCK_SIZE_X * 2 * gridDim.x * (idy               ) + idx + BLOCK_SIZE_X] = res.y;
    d_SLICE[BLOCK_SIZE_X * 2 * gridDim.x * (idy + BLOCK_SIZE_Y) + idx               ] = res.z;
    d_SLICE[BLOCK_SIZE_X * 2 * gridDim.x * (idy + BLOCK_SIZE_Y) + idx + BLOCK_SIZE_X] = res.w;
}

__global__ static void hst_cuda_nearest_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float2 res[2] = {0.f, 0.f, 0.f, 0.f};
    
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = 2 * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + 2 * blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + 2 * tidx;
    const int idy = bidy + 2 * tidy;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = idx + apos_off_x;
    const float y = idy + apos_off_y;


    __shared__  float   cache_subh[2*BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __shared__  float   cache_sin[BLOCK_SIZE_Y];
    __shared__  float   cache[BLOCK_SIZE_Y][3 * BLOCK_SIZE_X];
    
//#pragma unroll 8
    float projf = tidy + 0.5;
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE_Y) {
        float4 all = c_all[proje + tidy];
	int minh = (int)floor(all.z + bx * all.x - by * all.y + all.w);

	float subh = (all.z + x * all.x - minh);
	cache_subh[2*tidx][tidy] = subh;
	cache_subh[2*tidx+1][tidy] = subh + all.x;
	if (tidx == 0) cache_sin[tidy] = all.y;


#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;
	    cache[tidy][pos] = tex2D(tex_projes, minh + pos, proje + projf);
	}

	__syncthreads();

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE_Y; i++) {
	    int4 minh = {
		cache_subh[2 * tidx][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx][i] - (y + 1) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y + 1) * cache_sin[i],
	    };

	    res[0].x += cache[i][minh.x];
	    res[0].y += cache[i][minh.y];
	    res[1].x += cache[i][minh.z];
	    res[1].y += cache[i][minh.w];

	}


	__syncthreads();

    }

    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy    ) + idx) = res[0];
    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy + 1) + idx) = res[1];
}

__global__ static void hst_cuda_oversample4_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float2 res[2] = {0.f, 0.f, 0.f, 0.f};
    
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = 2 * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + 2 * blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + 2 * tidx;
    const int idy = bidy + 2 * tidy;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = idx + apos_off_x;
    const float y = idy + apos_off_y;


    __shared__  float   cache_subh[2*BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __shared__  float   cache_sin[BLOCK_SIZE_Y];
    __shared__  float   cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X];
    
//#pragma unroll 8
    float projf = tidy + 0.5;
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE_Y) {
        float4 all = c_all[proje + tidy];
	int minh = (int)floor(all.z + bx * all.x - by * all.y + all.w);

	float subh = 4 * (all.z + x * all.x - minh);
	cache_subh[2*tidx][tidy] = subh;
	cache_subh[2*tidx+1][tidy] = subh + 4 * all.x;
	if (tidx == 0) cache_sin[tidy] = 4 * all.y;


#pragma unroll 12
	for (int i = 0; i < 12; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;
	    cache[tidy][pos] = tex2D(tex_projes, minh + ((float)pos)/4, proje + projf);
	}

	__syncthreads();

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE_Y; i++) {
	    int4 minh = {
		cache_subh[2 * tidx][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx][i] - (y + 1) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y + 1) * cache_sin[i],
	    };

	    res[0].x += cache[i][minh.x];
	    res[0].y += cache[i][minh.y];
	    res[1].x += cache[i][minh.z];
	    res[1].y += cache[i][minh.w];

	}


	__syncthreads();

    }

    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy    ) + idx) = res[0];
    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy + 1) + idx) = res[1];
}
*/

// Both kernels have problems with pre sm2 builds. Can't tell.

#define PPT 4
// Iterations per line
#ifdef HST_MP_ASYMMETRIC
# define IPT (16/PPT)
#else /* HST_MP_ASYMMETRIC */
# ifdef HST_MP_MINIMAL
#  define IPT 4
# else /* HST_MP_MINIMAL */
#  define IPT 8
# endif /* HST_MP_MINIMAL */
#endif /* HST_MP_ASYMMETRIC */
#define PROJ_LINE (PPT * BLOCK_SIZE_X * 3 / 2)
#define WARP_LINES 32

__global__ static void hst_cuda_mplinear_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float res1[PPT][PPT] = {0.f};
#ifdef HST_MP_ASYMMETRIC
    float res2[PPT][PPT] = {0.f};
#endif /* HST_MP_ASYMMETRIC */
    
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = PPT * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + PPT * blockIdx.y * BLOCK_SIZE_Y;


    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    __shared__ float f_minh[128];
    __shared__ float2   cache[IPT * PROJ_LINE + 1];
    
    const int tbidy = tidy/(PPT/2);
    const int tbidx = tidy%(PPT/2);
    
#ifdef HST_OPTIMIZE_KEPLER
    const int wrapid = (tidy&1) * 16 + tidx;
#endif
    
#ifdef HST_MP_ASYMMETRIC
    const int sbidx = 8 * ((tidy / 2) % 2);
    const int sbidy = 8 * ((tidy / 2) / 2);
    
    const int stidx = tidx % 8;
    const int stidy = 2 * (2 * (tidy&1) + tidx / 8);
#else /* HST_MP_ASYMMETRIC */
    const int sbidx = 8 * ((tidy / 4) % 2);
    const int sbidy = 8 * ((tidy / 4) / 2);

    const int stidx = tidx % 8;
    const int stidy = 1 * (2 * (tidy&3) + tidx / 8);
#endif /* HST_MP_ASYMMETRIC */

    const int sidx = (sbidx + stidx);
    const int sidy = (sbidy + stidy);

    const int idx = bidx + sidx;
    const int idy = bidy + sidy;

    const float x = idx + apos_off_x;
    const float y = idy + apos_off_y;

    const float exp23 = exp2(23.f);
    
    const int ttidx = 16 * tidy + tidx;

    float projf = tbidy + 0.5f;
    
    const int num_proj_blocks = num_proj / 32;
    
    for (int proj_block=0; proj_block<num_proj_blocks; proj_block += 4) {
        const int proj = proj_block * 32;
        
        if (ttidx < 128) {
            float4 all = c_all[proj + ttidx];
            float minh = floor(all.z + bx * all.x - by * all.y + all.w);
            f_minh[ttidx] = minh;
        }
        
        __syncthreads();


#ifdef HST_OPTIMIZE_KEPLER
        float fminh[4];
#pragma unroll 4
        for (int i = 0; i < 4; i++) {
            fminh[i] = f_minh[32 * i + wrapid];
        }
#endif // HST_OPTIMIZE_KEPLER


        int max_proj = min(num_proj_blocks - proj_block, 4);

#pragma unroll 1
        for (int subproj32 = 0; subproj32 < max_proj; subproj32++) {
#pragma unroll 1
            for (int subproj = 0; subproj < (32/IPT); subproj++) {
                const int loop_proj = 32 * subproj32 + IPT * subproj;
                const int proje = proj + loop_proj;

//                float4 all = c_all[proje + tbidy];

#ifdef HST_SHFL_SUM
	        int minh = __shfl(fminh[subproj32], IPT * subproj + tbidy, 32);
#else //  HST_OPTIMIZE_KEPLER 
	        int minh = f_minh[loop_proj + tbidy];
#endif //  HST_OPTIMIZE_KEPLER         

#ifdef HST_MP_MINIMAL
                if (tidy < (PPT / 2) * IPT) {
#endif /* HST_MP_MINIMAL */
#pragma unroll 3
                    for (int i = 0; i < 3; i++) {
	                int pos = (i * (PPT / 2) + tbidx) * BLOCK_SIZE_X + tidx;
	        
	                cache[PROJ_LINE * tbidy + pos].x = tex2D(tex_projes, minh + pos, proje + projf);
                    }
                    
                        // without 'if' works without...
                    __threadfence_block();
	            //__syncthreads();

#pragma unroll 3
                    for (int i = 0; i < 3; i++) {
	                int pos = (i * (PPT / 2) + tbidx) * BLOCK_SIZE_X + tidx;
                        cache[tbidy * PROJ_LINE + pos].y = cache[tbidy * PROJ_LINE + pos + 1].x - cache[tbidy * PROJ_LINE + pos].x;
                    }
#ifdef HST_MP_MINIMAL
                }
#endif /* HST_MP_MINIMAL */
	        __syncthreads();

#pragma unroll 4 // IPT
	        for (int p = 0; p < IPT; p++) {
                    float4 all = c_all[proje + p];

#ifdef HST_SHFL_SUM
                    float minh = __shfl(fminh[subproj32], IPT * subproj + p, 32);
#else // HST_OPTIMIZE_KEPLER
                    float minh = f_minh[loop_proj + p];
#endif // HST_OPTIMIZE_KEPLER

                    float h = all.z + x * all.x - y * all.y - minh;


//                    char *cache_row = ((char*)(&cache)) + p * PROJ_LINE * sizeof(float2);


#pragma unroll 4 // PPT
	            for (int i = 0; i < PPT; i++) {
	                float subh = h;
	                h -= 16.f * all.y;
#pragma unroll 4 // PPT
	                for (int j = 0; j < PPT; j++) {
	                    float subh1 = subh;
#ifdef HST_MP_ASYMMETRIC
	                    float subh2 = subh - all.y;
#endif /* HST_MP_ASYMMETRIC */

#ifdef HST_MP_FANCY_ROUND
                            float fsubh1 = subh1 + exp23;
                            int idx1 = (*(int*)(&fsubh1)) - 0x4B000000;
                            fsubh1 = subh1 - (fsubh1 - exp23);

# ifdef HST_MP_ASYMMETRIC
                            float fsubh2 = subh2 + exp23;
                            int idx2 = (*(int*)(&fsubh2)) - 0x4B000000;
                            fsubh2 = subh2 - (fsubh2 - exp23);
# endif /* HST_MP_ASYMMETRIC */
#else /* HST_MP_FANCY_ROUND */
                            int idx1 = round(subh1);
                            float fsubh1 = subh1 - idx1;
# ifdef HST_MP_ASYMMETRIC
                            int idx2 = round(subh2);
                            float fsubh2 = subh2 - idx2;
# endif /* HST_MP_ASYMMETRIC */
#endif /* HST_MP_FANCY_ROUND */


	                    float2 c1 = cache[p * PROJ_LINE + idx1];
	                    res1[i][j] += c1.x + fsubh1 * c1.y;

#ifdef HST_MP_ASYMMETRIC
	                    float2 c2 = cache[p * PROJ_LINE + idx2];
	                    res2[i][j] += c2.x + fsubh2 * c2.y;
#endif /* HST_MP_ASYMMETRIC */

	                    subh += 16.f * all.x;
	                }
	            }

	        }
	    }
	    __syncthreads();
	}
    }

#pragma unroll 4
    for (int i = 0; i < PPT; i++) {
#pragma unroll 4
	for (int j = 0; j < PPT; j++) {
	    d_SLICE[BLOCK_SIZE_X * PPT * gridDim.x * (idy + i * BLOCK_SIZE_Y) + idx + j * BLOCK_SIZE_X] = res1[i][j];
#ifdef HST_MP_ASYMMETRIC
	    d_SLICE[BLOCK_SIZE_X * PPT * gridDim.x * (idy + i * BLOCK_SIZE_Y + 1) + idx + j * BLOCK_SIZE_X] = res2[i][j];
#endif /* HST_MP_ASYMMETRIC */
	}
    }
}


#undef PROJ_LINE
#define PROJ_LINE (2 * PPT * BLOCK_SIZE_X * 3)

__global__ static void hst_cuda_mpoversample4_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float res[PPT][PPT];
    
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = PPT * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + PPT * blockIdx.y * BLOCK_SIZE_Y;


    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    __shared__ float f_minh[256];
    __shared__ float cache[IPT * PROJ_LINE];
    
    const int tbidy = tidy/(16 / IPT);
    const int tbidx = tidy%(16 / IPT);
    
#ifdef HST_OPTIMIZE_KEPLER
    const int wrapid = (tidy&1) * 16 + tidx;
#endif
    
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;

    const int sidx = (4 * sbidx + stidx);
    const int sidy = (4 * sbidy + stidy);

    const int idx = bidx + sidx;
    const int idy = bidy + sidy;

    const float x = idx + apos_off_x;
    const float y = idy + apos_off_y;
    
    const int ttidx = 16 * tidy + tidx;

    float projf = tbidy + 0.5f;
    
    const int num_proj_blocks = num_proj / 32;
    
    for (int proj_block=0; proj_block<num_proj_blocks; proj_block += 8) {
        const int proj = proj_block * 32;
        
        float4 all = c_all[proj + ttidx];
        float minh = floor(all.z + bx * all.x - by * all.y + all.w);
        f_minh[ttidx] = minh;
        
        __syncthreads();

#ifdef HST_OPTIMIZE_KEPLER
        float fminh[8];

#pragma unroll 8
        for (int i = 0; i < 8; i++) {
            fminh[i] = f_minh[32 * i + wrapid];
        }
#endif //  HST_OPTIMIZE_KEPLER         

        int max_proj = min(num_proj_blocks - proj_block, 8);

#pragma unroll 1
        for (int subproj32 = 0; subproj32 < max_proj; subproj32++) {
#pragma unroll 1
            for (int subproj = 0; subproj < (32/IPT); subproj++) {
                const int loop_proj = 32 * subproj32 + IPT * subproj;
                const int proje = proj + loop_proj;

//                float4 all = c_all[proje + tbidy];

#ifdef HST_SHFL_SUM
	        int minh = __shfl(fminh[subproj32], IPT * subproj + tbidy, 32);
#else //  HST_OPTIMIZE_KEPLER 
	        int minh = f_minh[loop_proj + tbidy];
#endif //  HST_OPTIMIZE_KEPLER         


#pragma unroll 6
                for (int i = 0; i < 6; i++) {
	            int pos = (i * PPT + tbidx) * BLOCK_SIZE_X + tidx;
	            cache[PROJ_LINE * tbidy + pos] = tex2D(tex_projes, minh + 0.25f * pos, proje + projf);
                }


	        __syncthreads();

#pragma unroll 4 // IPT
	        for (int p = 0; p < IPT; p++) {
                    float4 all = 4 * c_all[proje + p];
                    all.y = -all.y;

#ifdef HST_SHFL_SUM
                    float minh = __shfl(fminh[subproj32], IPT * subproj + p, 32);
#else // HST_OPTIMIZE_KEPLER
                    float minh = f_minh[loop_proj + p];
#endif // HST_OPTIMIZE_KEPLER

                    float h = all.z + x * all.x + y * all.y - 4 * minh;

                    all.x *= 16.f;
                    all.y *= 16.f;

#pragma unroll 4 // PPT
	            for (int i = 0; i < PPT; i++) {
#pragma unroll 4 // PPT
	                for (int j = 0; j < PPT; j++) {
	                    float subh = h + i * all.y + j * all.x;
	                    res[i][j] += cache[p*PROJ_LINE + (int)subh];//cache[p * PROJ_LINE + (int)subh];
	                }
	            }


	        }
	    }
	    __syncthreads();

	}
    }

#pragma unroll 4
    for (int i = 0; i < PPT; i++) {
#pragma unroll 4
	for (int j = 0; j < PPT; j++) {
	    d_SLICE[BLOCK_SIZE_X * PPT * gridDim.x * (idy + i * BLOCK_SIZE_Y) + idx + j * BLOCK_SIZE_X] = res[i][j];
	}
    }
}
