__global__ static void hst_kepler_4slice_kernel(cudaTextureObject_t texptr, int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float4 *out = (float4*)d_SLICE;

    float h;
    float4 res[4][4] = {0};

#ifdef HST_OPTIMIZE_KEPLER
    __shared__ float4 buf[16][17]; // 64b for kepler
//    __shared__ float fill[48];
    __shared__ float4 fin[16][17];
#else // HST_OPTIMIZE_KEPLER
    __shared__ float4 buf[16][17]; // 32b for Fermi & GT200
//    __shared__ float4 fill[56];
    __shared__ float4 fin[16][17];
#endif // HST_OPTIMIZE_KEPLER

    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    const int sidx = tidx % 4;
    const int sidy = tidx / 4;

    const float x = bidx + sidx + apos_off_x;
    const float y = bidy + sidy + apos_off_y;

    const float projf = tidy + 0.5f;

    for (int proje=0; proje<num_proj; proje+=16) {
        const float4 all = c_all[proje+tidy];
        h = all.z + x * all.x - y * all.y;

#pragma unroll 4
        for (int i = 0; i < 4; i++) {
#pragma unroll 4
	    for (int j = 0; j < 4; j++) {
	        float subh = h + 4 * j * all.x - 4 * i * all.y;
		res[i][j] += tex2D<float4>(texptr, subh, proje + projf) ;
	    }
        }

    }

#pragma unroll 4
    for (int i = 0; i < 4; i++) {
#pragma unroll 4
	for (int j = 0; j < 4; j++) {
            buf[tidx][tidy] = res[i][j];

            __syncthreads();


#ifdef HST_OPTIMIZE_KEPLER
            float4 val = buf[tidy][tidx];
            for (int k=16; k>=1; k/=2) {
                val.x += __shfl_xor(val.x, k, 16);
                val.y += __shfl_xor(val.y, k, 16);
                val.z += __shfl_xor(val.z, k, 16);
                val.w += __shfl_xor(val.w, k, 16);
            }
#else // HST_OPTIMIZE_KEPLER
            volatile float4 *ptr = &buf[tidy][0];
            for (int k=8; k>=1; k/=2) {
                ptr[tidx].x += ptr[tidx+k].x;
                ptr[tidx].y += ptr[tidx+k].y;
                ptr[tidx].z += ptr[tidx+k].z;
                ptr[tidx].w += ptr[tidx+k].w;
    	    }
            float4 val;
            val.x = ptr[0].x;
            val.y = ptr[0].y;
            val.z = ptr[0].z;
            val.w = ptr[0].w;
#endif // HST_OPTIMIZE_KEPLER

            const int rx = 4 * j + tidy%4;
            const int ry = 4 * i + tidy/4;

            if (!tidx) {
                fin[ry][rx] = val;
            }

            __syncthreads();
        }
    }


    const int idx = bidx + tidx;
    const int idy = bidy + tidy;

    out[BLOCK_SIZE_X * gridDim.x * idy + idx] = fin[tidy][tidx];
}




#define SUMS 2
__global__ static void hst_cuda_kernel(cudaTextureObject_t texptr, int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    const int idx = blockIdx.x * BLOCK_SIZE_X + threadIdx.x;
    const int idy = blockIdx.y * BLOCK_SIZE_Y + threadIdx.y + batch;

    float h;
    float res=0;

    const float bx = idx + apos_off_x;
    const float by = idy + apos_off_y;

    for (int proje=0; proje<num_proj; proje++) {
        const float4 all = c_all[proje];
        h = all.z + bx * all.x - by * all.y;

# ifdef HST_CUDA_ARRAY
        res=res+tex2D<float>(texptr, h, proje + 0.5f) ;
# else /* HST_CUDA_ARRAY */
        res=res+tex2D(tex_projes, h, proje + 0.5f) ;
#endif /* HST_CUDA_ARRAY */

    }


/*
    float acc[SUMS] = {0};
    for (int proj=0; proj<num_proj; proj+=SUMS) {

	const float4 all1 = c_all[proj];
	const float4 all2 = c_all[proj + 1];

	float h1 = all1.z + bx * all1.x - by * all1.y;
	float h2 = all2.z + bx * all2.x - by * all2.y;

	acc[0] += tex2D(tex_projes, h1, proj + 0.5f);
	acc[1] += tex2D(tex_projes, h2, proj + 1.5f);
*/
/*
#pragma unroll 2
	for (int p = 0; p < SUMS; p++) {
	    int proje = proj + p;
	
	    const float4 all = c_all[proje];
	    h = all.z + bx * all.x - by * all.y;

	    acc[p] += tex2D(tex_projes, h, proje + 0.5f);
	}
*/
/*
    }

    for (int p = 0; p < SUMS; p++)
	res += acc[p];
*/

    d_SLICE[ BLOCK_SIZE_X*gridDim.x*idy + idx ] = res;
}


__global__ static void hst_kepler2_kernel(cudaTextureObject_t texptr, int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    float h;
    float res=0;

    const int sidx = 4 * (tidy % 4) + tidx % 4;
    const int sidy = 4 * (tidy / 4) + tidx / 4;

    const int idx = bidx + sidx;
    const int idy = bidy + sidy;

    const float bx = idx + apos_off_x;
    const float by = idy + apos_off_y;

#pragma unroll 1
    for (int proje=0; proje<num_proj; proje++) {
        const float4 all = c_all[proje];
        h = all.z + bx * all.x - by * all.y;

# ifdef HST_CUDA_ARRAY
        res=res+tex2D<float>(texptr, h, proje + 0.5f) ;
# else /* HST_CUDA_ARRAY */
        res=res+tex2D(tex_projes, h, proje + 0.5f) ;
#endif /* HST_CUDA_ARRAY */

    }

    d_SLICE[ BLOCK_SIZE_X*gridDim.x*idy + idx ] = res;
}



__global__ static void hst_kepler_kernel(cudaTextureObject_t texptr, int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float h;
    float res[4][4] = {0};

#ifdef HST_OPTIMIZE_KEPLER
    __shared__ float buf[16][18]; // 64b for kepler
    __shared__ float fill[48];
    __shared__ float fin[16][18];
#else // HST_OPTIMIZE_KEPLER
    __shared__ float buf[16][17]; // 32b for Fermi & GT200
    __shared__ float fill[56];
    __shared__ float fin[16][17];
#endif // HST_OPTIMIZE_KEPLER

    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    const int sidx = tidx % 4;
    const int sidy = tidx / 4;

    const float x = bidx + sidx + apos_off_x;
    const float y = bidy + sidy + apos_off_y;

    const float projf = tidx + 0.5f;//tidy + 0.5f;

    for (int proje=0; proje<num_proj; proje+=16) {
/*
        const float4 all = c_all[proje+tidy];
        h = all.z + x * all.x - y * all.y;
*/
        const float4 all = c_all[proje + tidy];
        buf[tidy][tidx] = all.z + x * all.x - y * all.y;

        __syncthreads();

        h = buf[tidx][tidy];

#pragma unroll 4
        for (int i = 0; i < 4; i++) {
#pragma unroll 4
	    for (int j = 0; j < 4; j++) {
	        float subh = h + 4 * j * all.x - 4 * i * all.y;
# ifdef HST_CUDA_ARRAY
                res[i][j] += tex2D<float>(texptr, subh, proje + projf);
# else /* HST_CUDA_ARRAY */
                res[i][j] += tex2D(tex_projes, subh, proje + projf);
#endif /* HST_CUDA_ARRAY */

	    }
        }

    }

    __syncthreads();


#pragma unroll 4
    for (int i = 0; i < 4; i++) {
#pragma unroll 4
	for (int j = 0; j < 4; j++) {
            buf[tidy][tidx] = res[i][j];

            __syncthreads();


#ifdef HST_OPTIMIZE_KEPLER
            float val = buf[tidy][tidx];
            for (int k=16; k>=1; k/=2)
                val += __shfl_xor(val, k, 16);
#else // HST_OPTIMIZE_KEPLER
            volatile float *ptr = &buf[tidy][0];
            for (int k=8; k>=1; k/=2) {
                ptr[tidx] += ptr[tidx+k];
            }

            float val = ptr[0];
#endif // HST_OPTIMIZE_KEPLER


            const int rx = 4 * j + tidy%4; //4 * j + tidy%4;
            const int ry = 4 * i + tidy/4; //4 * i + tidy/4;

            if (!tidx) {
                fin[ry][rx] = val;
            }

            __syncthreads();
        }
    }


    const int idx = bidx + tidx;
    const int idy = bidy + tidy;

    d_SLICE[BLOCK_SIZE_X * gridDim.x * idy + idx] = fin[tidy][tidx];
}

/* Newtex simple test */
__global__ static void hst_kepler_kernel(cudaTextureObject_t texptr, int num_proj, int num_bins, vfloat *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float h;
    vfloat res[4][4] = {0};

#ifdef HST_OPTIMIZE_KEPLER
    __shared__ float buf[16][18]; // 64b for kepler
    __shared__ float fill[48];
    __shared__ float fin[16][18];
#else // HST_OPTIMIZE_KEPLER
    __shared__ float buf[16][17]; // 32b for Fermi & GT200
    __shared__ float fill[56];
    __shared__ float fin[16][17];
#endif // HST_OPTIMIZE_KEPLER

    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;


    const int proj_minor = tidx / 4;
    const int xy_minor = tidx % 4;
    const int y_minor = xy_minor / 2;
    const int x_minor = xy_minor % 2;

    const int proj_major = tidy / 4;
    const int xy_major = tidy % 4;
    const int y_major = xy_major / 2;
    const int x_major = xy_major % 2;

    const int sidy = y_major * 2 + y_minor;
    const int sidx = x_major * 2 + x_minor;

    const float x = bidx + sidx + apos_off_x;
    const float y = bidy + sidy + apos_off_y;
    const float projf = 4 * proj_major + proj_minor + 0.5f;


    for (int proje=0; proje<num_proj; proje+=16) {
        const float4 all = c_all[proje+tidy];
        h = all.z + x * all.x - y * all.y;

#pragma unroll 4
        for (int i = 0; i < 4; i++) {
#pragma unroll 4
	    for (int j = 0; j < 4; j++) {
	        float subh = h + 4 * j * all.x - 4 * i * all.y;
	        res[i][j] += hst_tex(texptr, subh, proje + projf);
	    }
        }
    }


#pragma unroll 4
    for (int i = 0; i < 4; i++) {
#pragma unroll 4
	for (int j = 0; j < 4; j++) {
            buf[tidx][tidy] = res[i][j];

            __syncthreads();


#ifdef HST_OPTIMIZE_KEPLER
            float val = buf[tidy][tidx];
            for (int k=16; k>=1; k/=2)
                val += __shfl_xor(val, k, 16);
#else // HST_OPTIMIZE_KEPLER
            volatile float *ptr = &buf[tidy][0];
            for (int k=8; k>=1; k/=2)
                ptr[tidx] += ptr[tidx+k];
            float val = ptr[0];
#endif // HST_OPTIMIZE_KEPLER

            const int rx = 4 * j + tidy%4;
            const int ry = 4 * i + tidy/4;

            if (!tidx) {
                fin[ry][rx] = val;
            }

            __syncthreads();
        }
    }


    const int idx = bidx + tidx;
    const int idy = bidy + tidy;

    d_SLICE[BLOCK_SIZE_X * gridDim.x * idy + idx] = fin[tidy][tidx];
}

__global__ static void hst_kepler_4projection_kernel(cudaTextureObject_t texptr, int num_proj, int num_bins, vfloat *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float h;
    vfloat res[2][2] = {0};

    __shared__ vfloat buf[16][17]; // 32b for Fermi & GT200
    __shared__ vfloat fill[56];
    __shared__ vfloat fin[16][17];

    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    const int proj_minor = tidx / 4;
    const int xy_minor = tidx % 4;
    const int y_minor = xy_minor / 2;
    const int x_minor = xy_minor % 2;

    const int proj_major = 0; //tidy / 4;
    const int xy_major = tidy;
    const int y_major = xy_major / 4;
    const int x_major = xy_major % 4;

    const int proj = 4 * proj_major + proj_minor;
    const int sidy = y_major * 2 + y_minor;
    const int sidx = x_major * 2 + x_minor;

    const float x = bidx + sidx + apos_off_x;
    const float y = bidy + sidy + apos_off_y;
    const float projf = proj + 0.5f;

    for (int proje=0; proje<num_proj; proje+=4) {
        const float4 all = c_all[proje + proj];
        h = all.z + x * all.x - y * all.y;

#pragma unroll 2
        for (int i = 0; i < 2; i++) {
#pragma unroll 2
	    for (int j = 0; j < 2; j++) {
	        float subh = h + 8 * j * all.x - 8 * i * all.y;
	        res[i][j] += hst_tex(texptr, subh, proje + projf);
	    }
        }
    }


/*
#pragma unroll 2
    for (int i = 0; i < 2; i++) {
#pragma unroll 2
	for (int j = 0; j < 2; j++) {
            buf[tidx][tidy] = res[i][j];

            __syncthreads();

            for (int k=2; k>=1; k/=2)
                buf[tidy][4 * (tidx / 4) + tidx%4] += buf[tidy][4 * (tidx / 4) + tidx%4 + k];

	    vfloat val = buf[tidy][4 * (tidx / 4)];


//            volatile vfloat *ptr = &buf[tidy][4*(tidx/4)];
//            for (int k=2; k>=1; k/=2)
//                ptr[tidx%4] += ptr[tidx%4+k];
//            vfloat val = ptr[0];

//            for (int k=8; k>=1; k/=2) {
//                ptr[tidx].x += ptr[tidx+k].x;
//                ptr[tidx].y += ptr[tidx+k].y;
//    	    }
//    	    vfloat val;
//            val.x = ptr[0].x;
//            val.y = ptr[0].y;


            const int rx = 8 * j + 4 * (tidy%2) + tidx / 4;
            const int ry = 8 * i + tidy/2;

            if ((tidx%4) == 0) {
                fin[ry][rx] = val;
            }

            __syncthreads();
        }
    }
    const int idx = bidx + tidx;
    const int idy = bidy + tidy;

    d_SLICE[BLOCK_SIZE_X * gridDim.x * idy + idx] = fin[tidy][tidx];

*/

    const int idx = bidx + tidx;
    const int idy = bidy + tidy;

    d_SLICE[BLOCK_SIZE_X * gridDim.x * idy + idx] = res[0][0] + res[0][1] + res[1][0] + res[1][1];

}


#define TEXPPT 2

__global__ static 
#ifdef HST_SET_BOUNDS
__launch_bounds__(256, 8)//HST_NEWTEX_MIN_BLOCKS)
#endif /* HST_SET_BOUNDS */
#if defined(HST_FLOAT_LOOPS)
void hst_small_kepler_kernel(cudaTextureObject_t texptr, float num_proj, int num_bins, vfloat *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
#else
void hst_small_kepler_kernel(cudaTextureObject_t texptr, int num_proj, int num_bins, vfloat *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
#endif
    float h;
    vfloat res[TEXPPT][TEXPPT] = {0};

#if  (defined(HST_SHFL_SUM) && (SLICE_BLOCK == 1))
    __shared__ vfloat buf[16][20]; 
#else
    __shared__ vfloat buf[16][16]; 
#endif
    __shared__ vfloat fin[4 * TEXPPT][4 * TEXPPT];

    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    const int proj_minor = tidx / 4;
    const int xy_minor = tidx % 4;
    const int y_minor = xy_minor / 2;
    const int x_minor = xy_minor % 2;

    const int proj_major = tidy / 4;
    const int xy_major = tidy % 4;
    const int y_major = xy_major / 2;
    const int x_major = xy_major % 2;

    const int proj =  4 * proj_major + proj_minor;
    const int sidy = y_major * 2 + y_minor;
    const int sidx = x_major * 2 + x_minor;

    const float x = 0.25f * (bidx + sidx + apos_off_x);	// we need increase sin/cos by 4x in system code
    const float y = 0.25f * (bidy + sidy + apos_off_y);

#if defined(HST_FLOAT_LOOPS)
    for (float projf = proj + 0.5f; projf < num_proj; projf += 16) {
//        float fidx = projf + exp2(23.f); const int idx = (*(int*)(&fidx)) - 0x4B000000;
        const int proje = (int)projf;
#else
    for (int proje = proj; proje < num_proj; proje += 16) {
	const float projf = proje + 0.5f;
#endif
        const float4 all = c_all[proje];		// (int)projf (or fancy round) to avoid register spills
        h = all.z + x * all.x - y * all.y;

	res[0][0] += hst_tex(texptr, h, projf);
	res[0][1] += hst_tex(texptr, h + all.x, projf);
	res[1][0] += hst_tex(texptr, h - all.y, projf);
	res[1][1] += hst_tex(texptr, h + all.x - all.y, projf);

//#pragma unroll 2
//        for (int i = 0; i < TEXPPT; i++) {
//#pragma unroll 2
//	    for (int j = 0; j < TEXPPT; j++) {
//	        float subh = h + /*4 */ j * all.x - /*4 */ i * all.y;
//	        res[i][j] += hst_tex(texptr, subh, projf);

//	    }
//        }
    }


#pragma unroll 2
    for (int i = 0; i < TEXPPT; i++) {
#pragma unroll 2
	for (int j = 0; j < TEXPPT; j++) {
            buf[4 * sidy + sidx][proj] = res[i][j];

            __syncthreads();

#ifdef HST_SHFL_SUM
            vfloat val = buf[tidy][tidx];
//#pragma unroll 5
            for (int k=16; k>=1; k/=2) 
        	shfl_sum(val, k);
# else // HST_SHFL_SUM
//#pragma unroll 5
            for (int k=8; k>=1; k/=2) {
                buf[tidy][tidx] += buf[tidy][tidx+k];
            	    // We need it instead of volatile. Could be faster if buf declared volatile
                __syncthreads();
            }
            vfloat val = buf[tidy][0];
#endif //  HST_SHFL_SUM

            if (!tidx) {
	        const int rx = 4 * j + tidy%4;
        	const int ry = 4 * i + tidy/4;

		const int idx = bidx + rx;
		const int idy = bidy + ry;
		d_SLICE[(BLOCK_SIZE_X * TEXPPT / 4)  * gridDim.x * idy + idx] = val;
//                fin[ry][rx] = val;
            }

            __syncthreads();
        }
    }

/*
    const int idx = bidx + tidx;
    const int idy = bidy + tidy;

    d_SLICE[BLOCK_SIZE_X * gridDim.x * idy + idx] = fin[tidy][tidx];
*/
}

