#if SLICE_BLOCK == 4
# define vfloat float4
# define tex2d(a,b,c) read_imagef(a, b, c)
# define SET_SLICE(slice, idx, idy, sum) ({ slice[idy*slice_width + idx] = sum.x; slice[slice_size + idy*slice_width + idx] = sum.y; slice[2 * slice_size + idy*slice_width + idx] = sum.z; slice[3 * slice_size + idy*slice_width + idx] = sum.w; })
#elif SLICE_BLOCK == 2
# define vfloat float2
# define tex2d(a,b,c) read_imagef(a, b, c).xy
# define SET_SLICE(slice, idx, idy, sum) ({ slice[idy*slice_width + idx] = sum.x; slice[slice_size + idy*slice_width + idx] = sum.y; })
#else
# define vfloat float
# define tex2d(a,b,c) read_imagef(a, b, c).x
# define SET_SLICE(slice, idx, idy, sum) ({ slice[idy*slice_width + idx] = sum; })
#endif


inline vfloat interpolate_lin(__global vfloat *sinogram, float pos) {
    int idx = floor(pos);
    float delta = pos - idx;
    vfloat v1 = sinogram[idx];
    return v1 + delta * (sinogram[idx + 1] - v1);
}

inline vfloat interpolate_nn(__global vfloat *sinogram, float pos) {
    return  sinogram[(int)(pos)];
}

#if OVERSAMPLE == 1
# define interpolate interpolate_nn
#else
# define interpolate interpolate_lin
#endif


__kernel void hst_backproject_cpu_orig(const int num_proj,
    const int num_bins,
    const float off_x,
    const float off_y,
    __global vfloat *sinogram,
    __global float *slice,
    __constant float4 *all_table)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);

    const int slice_width = get_global_size(0);
    const int slice_size = get_global_size(1) * slice_width;

    // TODO: maybe use float4 for optimal vectorization on Intel CPUs
    float2 b = (float2) (idx + off_x, idy + off_y);
    vfloat sum = 0.0f;

//#pragma unroll 16
    for(int proj = 0; proj < num_proj; proj++) {
        float2 s = (float2) (all_table[proj].x, all_table[proj].y);
        s = s * b;

	float pos = proj * num_bins + all_table[proj].z + s.x - s.y;
	sum += interpolate(sinogram, pos);

//        sum += sinogram[(int)(proj * num_bins + corr + s.x - s.y)];
    }

    SET_SLICE(slice, idx, idy, sum);
}



#if PPT > 1
# define YTEXSIZE (PPT * PPT)
__kernel void hst_backproject_cpu(const int num_proj,
    const int num_bins,
    const float off_x,
    const float off_y,
    __global vfloat *sinogram,
    __global float *slice,
    __constant float4 *all_table)
{
//    const int idx = get_global_id(0);
//    const int idy = get_global_id(1);

    const int bx = get_group_id(0) * get_local_size(0);
    const int by = get_group_id(1) * get_local_size(1);

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int sidx = YTEXSIZE * (tidx % (get_local_size(0) / PPT));
    const int sidy = PPT * tidy + tidx / (get_local_size(1) / PPT);


    const int idx = PPT * bx + sidx;
    const int idy = PPT * by + sidy;

    const int slice_width = get_global_size(0) * PPT;
    const int slice_size = get_global_size(1) * slice_width * PPT;

    // TODO: maybe use float4 for optimal vectorization on Intel CPUs
    float4 b = (float4) (idx + off_x, idy + off_y, idx + off_x, idy + off_y);

    vfloat sum[PPT * PPT] = {0.0f};

//#pragma unroll 4
    for(int proj = 0; proj < num_proj; proj += 2) {
        float4 s = (float4) (all_table[proj].x, all_table[proj].y, all_table[proj + 1].x, all_table[proj + 1].y);
        s = s * b;

	float pos = proj * num_bins + all_table[proj].z + s.x - s.y;
#pragma unroll
	for (int i = 0; i < YTEXSIZE; i++) {
	    sum[i] += interpolate(sinogram, pos + i * all_table[proj].x);
	}

	pos = (proj + 1) * num_bins + all_table[proj + 1].z + s.z - s.w;
#pragma unroll
	for (int i = 0; i < YTEXSIZE; i++) {
	    sum[i] += interpolate(sinogram, pos + i * all_table[proj + 1].x);
	}

//	barrier(CLK_LOCAL_MEM_FENCE);
    }

#pragma unroll
    for (int i = 0; i < YTEXSIZE; i++) {
	vfloat res = sum[i];
	SET_SLICE(slice, idx + i, idy, res);
    }
}
#else
__kernel void hst_backproject_cpu(const int num_proj,
    const int num_bins,
    const float off_x,
    const float off_y,
    __global vfloat *sinogram,
    __global float *slice,
    __constant float4 *all_table)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);

    const int slice_width = get_global_size(0);
    const int slice_size = get_global_size(1) * slice_width;

    // TODO: maybe use float4 for optimal vectorization on Intel CPUs
    float4 b = (float4) (idx + off_x, idy + off_y, idx + off_x, idy + off_y);

    vfloat sum = 0.0f;

//#pragma unroll 4
    for(int proj = 0; proj < num_proj; proj += 2) {
        float4 s = (float4) (all_table[proj].x, all_table[proj].y, all_table[proj + 1].x, all_table[proj + 1].y);
        s = s * b;

	float pos = proj * num_bins + all_table[proj].z + s.x - s.y;
	sum += interpolate(sinogram, pos);

	pos = (proj + 1) * num_bins + all_table[proj + 1].z + s.z - s.w;
	sum += interpolate(sinogram, pos);
    }

    SET_SLICE(slice, idx, idy, sum);
}
#endif

