/*
 * The PyHST program is Copyright (C) 2002-2011 of the
 * European Synchrotron Radiation Facility (ESRF) and
 * Karlsruhe Institute of Technology (KIT).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * hst is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

static const char *hst_opencl_kernels_source = 
#include "hst_opencl_kernels.cl.h"

static const char *hst_opencl_alu_source =
#include "alu.cl.h"

static const char *hst_opencl_cpu_source =
#include "cpu.cl.h"

static const char *hst_opencl_newtex_source =
#include "newtex.cl.h"

static const char **hst_opencl_sources[] = { &hst_opencl_kernels_source, &hst_opencl_alu_source, &hst_opencl_cpu_source, &hst_opencl_newtex_source, NULL };

//#define HST_PREFER_ORIG
#ifndef HST_PREFER_ORIG
//# define HST_MULTI_POINT_RECONSTRUCTION
//# define HST_PREFER_ALU
#endif

#define HST_MAX_KERNEL_VARIANTS 5


struct HSTKernelDescriptionT {
    const char *kernel;
    size_t ppt;
    size_t bp_block;
    size_t shared_memory;
};
typedef struct HSTKernelDescriptionT HSTKernelDescription;

static const HSTKernelDescription HST_KERNEL_DESCRIPTION[] = {
    {"hst_backproject_gpu_oversample4",		1,	BLOCK_SIZE,	(7 * BLOCK_SIZE * BLOCK_SIZE + 2 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_gpu_linear", 		1,	BLOCK_SIZE,	(8 * BLOCK_SIZE * BLOCK_SIZE + 3 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_gpu_nearest", 		1,	BLOCK_SIZE,	(5 * BLOCK_SIZE * BLOCK_SIZE + 3 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_fermi_oversample4", 	2,	BLOCK_SIZE,	(14 * BLOCK_SIZE * BLOCK_SIZE + 3 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_fermi_linear", 		2,	BLOCK_SIZE,	(8 * BLOCK_SIZE * BLOCK_SIZE + 5 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_fermi_nearest", 		2,	BLOCK_SIZE,	(5 * BLOCK_SIZE * BLOCK_SIZE + 3 * BLOCK_SIZE) * sizeof(float)},
    {"hst_backproject_direct_nearest", 		2,	BLOCK_SIZE,	(3 * BLOCK_SIZE + 2) * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_amd_direct_nearest", 	2,	BLOCK_SIZE,	(3 * BLOCK_SIZE + 2) * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_direct_oversample4", 	2,	BLOCK_SIZE,	(12 * BLOCK_SIZE + 1) * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_amd_direct_oversample4", 	2,	BLOCK_SIZE,	(12 * BLOCK_SIZE + 2) * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_direct_linear", 		2,	BLOCK_SIZE,	(6 * BLOCK_SIZE + 2) * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_amd_direct_linear", 	2,	BLOCK_SIZE,	(6 * BLOCK_SIZE + 2) * BLOCK_SIZE * sizeof(float)},
//    {"hst_backproject_6ppt_oversample4", 5 * BLOCK_SIZE * BLOCK_SIZE * sizeof(float)},
    {"hst_backproject_6ppt_oversample4", 	6,	8,		5 * 8 * 8 * sizeof(float)},
    {"hst_backproject_8ppt_oversample4", 	8,	8,		6 * 8 * 8 * sizeof(float)},
    {"hst_backproject_8ppt_linear", 		8,	8,		2 * (2 * 8 * 8 + 1) * sizeof(float)},
/*    {"hst_backproject_8ppt_vliw_oversample4", 6 * 8 * 8 * sizeof(float)},
    {"hst_backproject_8ppt_vliw_linear", 2 * (2 * 8 * 8 + 1) * sizeof(float)},*/
//    {"hst_backproject_kepler_tex", (2 * (16 * 18) + 56) * sizeof(float)},
//    {"hst_backproject_gcn_tex", (2 * (16 * 18) + 56) * sizeof(float)},
    { NULL, 0}
};


/* NOTE: the following kernel names MUST match the preceeding IDs */
static const char *HST_KERNEL_NAMES[HST_DEVICE_NUMBER][HST_KERNEL_NUMBER][HST_MAX_KERNEL_VARIANTS] = {
    { // Default
	{"hst_filter"},
#if defined(HST_PREFER_ORIG)
	{"hst_backproject_cpu_orig"},
#elif defined(HST_MULTI_POINT_RECONSTRUCTION)
# ifdef HST_PREFER_ALU
	{"alu"},
# else
	{"hst_backproject_cpu"},
# endif
#else
	{"hst_backproject_cpu_orig"},
#endif
//	{"hst_backproject_amd_direct_linear"},
	{"hst_spread"},
	{"hst_pack"},
	{"hst_normalize_idft"}
    },{ // GPU/Default
	{},
#if defined(HST_PREFER_ORIG)
	{"hst_backproject_gpu"},
#elif defined(HST_MULTI_POINT_RECONSTRUCTION)
	{"alu"},
#else
	{"newtex"},
#endif
//	{"hst_backproject_kepler_tex"},
	{},
	{},
	{}
#if SLICE_BLOCK == 0
    },{ // GT200
	{},
#ifdef HST_OPENCL_CONST_HACK
	{"hst_backproject_gpu_const"},
#else /* HST_OPENCL_CONST_HACK */
	{},
#endif /* HST_OPENCL_CONST_HACK */
	{},
	{},
	{}
    },{ // FERMI (Fermi kernels are expected to process 4 pixels per thread)
	{},
#ifdef HST_MULTI_POINT_RECONSTRUCTION
	{"hst_backproject_fermi_linear", "hst_backproject_fermi_nearest", "hst_backproject_fermi_oversample4"},
//	{"hst_backproject_direct_linear", "hst_backproject_direct_nearest", "hst_backproject_direct_oversample4"},
#else /* HST_MULTI_POINT_RECONSTRUCTION */
	{"hst_backproject_gpu_linear", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
#endif /* HST_MULTI_POINT_RECONSTRUCTION */
	{},
	{},
	{}
    },{ // Kepler
	{},
#ifdef HST_MULTI_POINT_RECONSTRUCTION
	{"hst_backproject_6ppt_tex", "hst_backproject_fermi_nearest", "hst_backproject_6ppt_oversample4"},
#else /* HST_MULTI_POINT_RECONSTRUCTION */
	{"hst_backproject_gpu_linear", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
#endif /* HST_MULTI_POINT_RECONSTRUCTION */
	{},
	{},
	{}
    },{ // Pascal
	{},
#ifdef HST_MULTI_POINT_RECONSTRUCTION
	{"hst_backproject_6ppt_tex", "hst_backproject_fermi_nearest", "hst_backproject_6ppt_oversample4"},
#else /* HST_MULTI_POINT_RECONSTRUCTION */
	{"hst_backproject_gpu_linear", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
#endif /* HST_MULTI_POINT_RECONSTRUCTION */
	{},
	{},
	{}
    },{ // Cypress
	{},
#ifdef HST_MULTI_POINT_RECONSTRUCTION
	{"alu", "alu", "alu"},
//	{"hst_backproject_direct_linear", "hst_backproject_amd_direct_nearest", "hst_backproject_amd_direct_oversample4"},
////	{"hst_backproject_4ppt_tex", "hst_backproject_amd_direct_nearest", "hst_backproject_direct_oversample4"},
//	{"hst_backproject_4ppt_tex", "hst_backproject_amd_direct_nearest", "hst_backproject_8ppt_oversample4"},
//	{"hst_backproject_6ppt_vliw_tex", "hst_backproject_amd_direct_nearest", "hst_backproject_8ppt_vliw_oversample4"},
#else /* HST_MULTI_POINT_RECONSTRUCTION */
	{"hst_backproject_kepler_tex", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
//	{"hst_backproject_gpu", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},

#endif /* HST_MULTI_POINT_RECONSTRUCTION */
	{},
	{},
	{}
    },{ // GCN
	{},
#ifdef HST_MULTI_POINT_RECONSTRUCTION
	{"alu", "alu", "alu"},
//	{"hst_backproject_direct_linear", "hst_backproject_amd_direct_nearest", "hst_backproject_direct_oversample4"},
//	{"hst_backproject_amd_direct_linear", "hst_backproject_amd_direct_nearest", "hst_backproject_amd_direct_oversample4"},
//	{"hst_backproject_fermi_linear", "hst_backproject_fermi_nearest", "hst_backproject_fermi_oversample4"},
#else /* HST_MULTI_POINT_RECONSTRUCTION */
//	{"hst_backproject_gpu", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
	{"hst_backproject_kepler_tex", "hst_backproject_gpu_nearest", "hst_backproject_gpu_oversample4"},
#endif /* HST_MULTI_POINT_RECONSTRUCTION */
	{},
	{},
	{}
    },{ // AMD/CPU
	{},
	{},
	{},
	{},
	{}
    },{ // Intel/CPU
	{},
	{},
	{},
	{},
	{}
#endif
    }
};

inline const HSTKernelDescription *hst_opencl_resolve_kernel_description(OpenCLContext *ctx, const char *name) {
    int i;
    
    for (i = 0; HST_KERNEL_DESCRIPTION[i].kernel; i++) {
	if (!strcmp(HST_KERNEL_DESCRIPTION[i].kernel, name)) return HST_KERNEL_DESCRIPTION + i;
    }
    
    return NULL;
}


inline hst_device_type hst_opencl_resolve_kernel_architecture(OpenCLContext *ctx, hst_kernel_type kernel_id) {
    if (HST_KERNEL_NAMES[ctx->arch][kernel_id][0]) return ctx->arch;
    if ((ctx->device_type == CL_DEVICE_TYPE_GPU)&&(HST_KERNEL_NAMES[HST_DEVICE_GPU][kernel_id][0])) return HST_DEVICE_GPU;
    return HST_DEVICE_UNKNOWN;
}

static const char *hst_opencl_resolve_kernel(OpenCLContext *ctx, hst_kernel_type kernel_id, int variant) {
    int i;
    hst_device_type arch = hst_opencl_resolve_kernel_architecture(ctx, kernel_id);
//    printf("Architecture: %i (device %i==%i arch %i/kernel %i/%p)\n", arch, ctx->device_type, CL_DEVICE_TYPE_GPU, ctx->arch, kernel_id, HST_KERNEL_NAMES[HST_DEVICE_GPU][kernel_id][0]);
    if (variant >= HST_MAX_KERNEL_VARIANTS) variant = HST_MAX_KERNEL_VARIANTS - 1;
    
    for (i = 0; i < variant; i++) {
	if (!HST_KERNEL_NAMES[arch][kernel_id][i+1]) break;
    }

    return HST_KERNEL_NAMES[arch][kernel_id][i];
}

static char* hst_opencl_load_program(const char* filename)
{
    FILE *fp = fopen(filename, "r");
    if (fp == NULL) {
        pyhst_error("Error: Couldn't open %s\n", filename);
        return NULL;
    }
    fseek (fp , 0 , SEEK_END);
    const size_t len = ftell(fp);
    rewind(fp);
    char *buffer;
    buffer = (char*) malloc (len * sizeof(char));
    if (buffer == NULL)
        return NULL;
    size_t buflen = fread(buffer, 1, len, fp);
    fclose(fp);
    if (buflen != len) {
        pyhst_error("Error: Couldn't read program %s from disk\n", filename);
        free(buffer);
        return NULL;
    }
    return buffer;
}


static int hst_opencl_load_kernels(OpenCLContext *ctx, int *variants) {
    int i;
    int err;
    hst_kernel_type kernel_id;
    
    HSTSetup *setup = ((HSTReconstructorContext*)ctx)->setup;
    int num_proj = setup->num_projections;

    char *arch;

    switch(ctx->arch) {
	case HST_DEVICE_GPU:
	    arch = "-D GPU";
	    break;
	case HST_DEVICE_CYPRESS:
	    arch = "-D GPU -D GPU_AMD -D GPU_VLIW5";
	    break;
	case HST_DEVICE_GCN:
	    arch = "-D GPU -D GPU_AMD -D GPU_GCN";
	    break;
	case HST_DEVICE_GT200:
	    arch = "-D GPU -D GPU_NV -D GPU_GT200";
	    break;
	case HST_DEVICE_FERMI:
	    arch = "-D GPU -D GPU_NV -D GPU_FERMI";
	    break;
	case HST_DEVICE_KEPLER:
	    arch = "-D GPU -D GPU_NV -D GPU_KEPLER";
	    break;
	case HST_DEVICE_PASCAL:
	    arch = "-D GPU -D GPU_NV -D GPU_PASCAL";
	    break;
	case HST_DEVICE_AMD_CPU:
	case HST_DEVICE_INTEL_CPU:
	    arch = "-D CPU";
	    break;
	default:
	    arch = "";
    }		

    char *build_flags = alloca(sizeof(HST_OPENCL_BUILD_FLAGS) + 1024);
    sprintf(build_flags, "%s -D SLICE_BLOCK=%u -D BLOCK_SIZE=%u -D BP_BLOCK=%u -D PPT=%u -D ASSYMETRY=%u -D %s=%u %s %s", arch, SLICE_BLOCK, BLOCK_SIZE, ctx->bp_block, ctx->ppt, ctx->assymetry,
	(setup->oversampling?"OVERSAMPLE":"NON_OVERSAMPLE"), setup->oversampling,
#ifdef HST_TEXTURE_READS
	"-D HST_TEXTURE_READS",
#else
	"",
#endif
	HST_OPENCL_BUILD_FLAGS
    );

    printf("Build flags: %s\n", build_flags);

    for (i = 0;  hst_opencl_sources[i]; i++) {
	const char *source = *(hst_opencl_sources[i]);

	const char *source_end = strstr(source, "//HST_OPENCL_CONST_HACK");
	if (source_end) {
	    size_t source_len = source_end - source;
	    char *source_buf = malloc(source_len + 1);
	    check_alloc(source_buf, "source_buf");
	    memcpy(source_buf, source, source_len);
	    source_buf[source_len] = 0;
        
	    ctx->program[i] = clCreateProgramWithSource(ctx->clctx, 1, (const char**)&source_buf, NULL, &err);
	    free(source_buf);
	} else {
	    ctx->program[i] = clCreateProgramWithSource(ctx->clctx, 1, &source, NULL, &err);
	}

	CL_CHECK_ERROR(err);

	err = clBuildProgram(ctx->program[i], 0, NULL, build_flags, NULL, NULL);
	if (err != CL_SUCCESS) {
	    const int LOG_SIZE = 4096;
	    char* log = (char*) malloc(LOG_SIZE * sizeof(char));
	    CL_CHECK_ERROR_EX(clGetProgramBuildInfo(ctx->program[i], ctx->device_id, CL_PROGRAM_BUILD_LOG, LOG_SIZE, (void*) log, NULL));
    	    pyhst_error("OpenCL Build Log\n---------------\n%s", log);
    	    free(log);
    	    return 255;
	}
    }

#ifdef HST_OPENCL_SAVE_PTX    
    void *binary;
    size_t binary_size;
    size_t real_size;

    err = clGetProgramInfo(ctx->program, CL_PROGRAM_BINARY_SIZES, sizeof(size_t), &binary_size, &real_size);
    if (err == CL_SUCCESS) binary = malloc(binary_size);
    if ((binary)&&(err == CL_SUCCESS)) {
	FILE *f;
	err = clGetProgramInfo(ctx->program[0], CL_PROGRAM_BINARIES, sizeof(void*), &binary, &real_size);
	if (err == CL_SUCCESS) f = fopen(HST_OPENCL_SAVE_PTX, "wb");
	else f = NULL;
	
	if (f) {
	    (void)fwrite(binary, 1, binary_size, f);
	    fclose(f);
	}
	free(binary);
    }
#endif /* HST_OPENCL_SAVE_PTX */

    for (kernel_id = 0; kernel_id < HST_KERNEL_NUMBER; kernel_id++) {
	cl_program prog;
	
	const char *kernel_name = hst_opencl_resolve_kernel(ctx, kernel_id, variants?variants[kernel_id]:0);
	printf("%s: loading the kernel %s\n", ctx->device_name, kernel_name);

	if (strstr(kernel_name, "alu")) prog = ctx->program[HST_PROGRAM_ALU];
	else if (strstr(kernel_name, "cpu")) prog = ctx->program[HST_PROGRAM_CPU];
	else if (strstr(kernel_name, "newtex")) prog = ctx->program[HST_PROGRAM_NEWTEX];
	else prog = ctx->program[0];

	ctx->kernel[kernel_id] = clCreateKernel(prog, kernel_name, &err);
        CL_CHECK_ERROR_EX(err);
        if (err != CL_SUCCESS) {
            pyhst_error("Undefined kernel name %s", kernel_name);
	    continue;
	}
	
	const HSTKernelDescription *d = hst_opencl_resolve_kernel_description(ctx, kernel_name);
	if ((d)&&(d->shared_memory)) {
	    cl_uint args;
	    err = clGetKernelInfo(ctx->kernel[kernel_id], CL_KERNEL_NUM_ARGS, sizeof(cl_uint), &args, NULL);
	    if (err) {
	        pyhst_error("Unable to determine number of arguments for kernel %s", kernel_name);
		continue;
	    }
	    
	    err = clSetKernelArg(ctx->kernel[kernel_id], args - 1, d->shared_memory, NULL);
	    if (err) {
	        pyhst_error("Unable to set shared memory argument for kernel %s", kernel_name);
		continue;
	    }
/*	    
	    if (!strcmp(kernel_name, "hst_backproject_direct_oversample4")) {
	        err = clSetKernelArg(ctx->kernel[kernel_id], args - 2, d->shared_memory, NULL);
	    }

	    if (err) {
	        pyhst_error("Unable to set shared memory argument for kernel %s", kernel_name);
		continue;
	    }
*/
	}
    }
    
    return err;
}
